@@ -16,19 +16,153 @@ As a nod to runtime efficiency, offsets useful for a displacement table are prec
 and provided for later.
 """
 
-import collections
-from . import foundation
+import collections, bisect
+from typing import List, Optional, Tuple, Sequence, Set
+from . import foundation, expansion
 
+VERBOSE = False
+
+###
+###  A few convenient abbreviations: Skip them on first reading.
+###
 
 def most_common(row, *, default=0):
 	""" Return the most-common element in the array; break ties arbitrarily. """
 	C = collections.Counter(row)
 	return C.most_common(1)[0][0] if C else default
 
-def is_homogeneous(row):
-	return len(row) < 2 or all(r == row[0] for r in row)
+def is_homogeneous(row): return len(row) < 2 or all(r == row[0] for r in row)
+
+def heavy_rows(matrix):
+	""" Return the indices of rows that are more-than-half full. Fullness == Truthiness. """
+	return [r for r, row in enumerate(matrix) if sum(map(bool, row)) * 2 > len(row)]
+
+def lighten(row):
+	""" Preserve `None` (as don't-care); logical-NOT the rest. """
+	return tuple(None if x is None else int(not x) for x in row)
+
+def measure_approximate_cost(structure):
+	""" Various bits estimate the size of the structures they return. This makes that consistent. """
+	if isinstance(structure, (list, tuple)): return 1 + sum(map(measure_approximate_cost, structure))
+	elif isinstance(structure, dict): return len(structure) + sum(map(measure_approximate_cost, structure.values()))
+	elif isinstance(structure, int) or structure is None: return 1
+	else: assert False, type(structure)
+
+###
+###  Leitmotifs: Recognizable themes that get used over and over...
+###
+
+def find_row_equivalence(matrix, do_not_care) -> Tuple[List[int], List[List]]:
+	"""
+	:param matrix: A rank-two tensor. Iterable of rows.
+	:param do_not_care: Cells with this value may be assigned any convenient value for the purpose.
+	"""
+	index, classes = [], []
+	for row in matrix:
+		for class_id, candidate_class in enumerate(classes):
+			if all(a == b or do_not_care in (a, b) for a, b in zip(row, candidate_class)):
+				index.append(class_id)
+				for c, value in enumerate(row):
+					if value != do_not_care: candidate_class[c] = value
+				break
+		else:
+			index.append(foundation.allocate(classes, list(row)))
+	return index, classes
+
+def find_column_equivalence(matrix, do_not_care) -> Tuple[List[int], List[Sequence]]:
+	""" Adapt find_row_equivalence (above) to work on columns instead of rows. """
+	index, classes = find_row_equivalence(zip(*matrix), do_not_care)
+	return index, list(zip(*classes))
+
+def first_fit_decreasing(indices:list, *, allow_negative:bool) -> Tuple[List[int], int] :
+	"""
+	(Aho and Ulman [2]) and (Ziegler [7]) advocate this scheme...
+	This function finds a set of displacements such that no two index[i][j]+displacement[i] are the same.
 	
-def compress_dfa_matrix(*, initial:dict, matrix:list, final:dict) -> dict:
+	This is part of a strategy for compact representation of a sparse matrix
+	with fast O(1) read performance. The input to THIS function is a statement
+	about which cells in the original matrix contain non-trivial data:
+	`col in indices[row]` exactly when `matrix[row][column]` is non-trivial.
+
+	The result can be used to populate a displacement table, which is a simple kind of perfect-hash.
+		* A hash is "perfect" when it requires at most one probe to determine
+		the presence or absence of a key; in general these must be precomputed.
+		* A hash is "minimal" when (in the lingo of hashing) all the buckets are used.
+		* The holy grail of "minimal perfect hashing" generally requires a more complex function.
+	
+	Rows are considered in decreasing order of density, so if Zipf's law applies then the result
+	will tend to be packed fairly densely: i.e. minimal or nearly so. This tends to be true of the
+	way this particular function gets used in this library.
+	
+	The size of the necessary resulting vector is also returned: it would have to be determined again
+	in every reasonable case otherwise.
+	
+	A final note: empty input rows are assigned the offset equal to the size of the implied check-vector.
+	"""
+	used = set()
+	def first_fit(row: List[int]) -> int:
+		offset = 0 - min(row) if allow_negative else 0
+		while any(r + offset in used for r in row): offset += 1
+		used.update(r + offset for r in row)
+		return offset
+	
+	displacements = [0] * len(indices) # Allocate space; it will be filled out-of-order.
+	populations = list(map(len, indices))
+	schedule = iter(foundation.grade(populations, descending=True))
+	for i in schedule:
+		if indices[i]: displacements[i] = first_fit(indices[i])
+		else: break # All remaining should be set == size of implied check vector.
+	size = max(used)+1 if used else 0
+	for i in schedule: displacements[i] = size
+	if VERBOSE: print("First-Fit Decreasing placed %d entries among %d positions."%(len(used), size))
+	return displacements, size
+
+def encode_boolean_field_as_offset_check(matrix:Sequence[Sequence[Optional[bool]]]):
+	"""
+	This is used in both the scanner and parser tables for keeping track of the error plane.
+	This function does not worry about equivalence classification.
+	It returns a pair of vectors such that:
+		check[offset[row]+column] == column
+	is equivalent to:
+		bool(matrix[row][column])
+	The returned structure is smaller when it has fewer `True` entries to encode.
+	"""
+	ones = [set(c for c,flag in enumerate(row) if flag) for row in matrix]
+	offset, size = first_fit_decreasing(ones, allow_negative=True)
+	check = [-1] * size
+	for r, (base, cs) in enumerate(zip(offset, ones)):
+		for c in cs: check[base+c] = r
+	return offset, check
+
+def encode_displacement_function(exceptions:list):
+	"""
+	This builds a typical textbook-style displacement table: a simple kind of perfect-hash,
+	which may not be minimal but it's usually decently close.
+	:param exceptions: a list of dictionaries; keys are column numbers, values are destination state numbers.
+	No attempt is made to coalesce similar or identical rows here:
+		(1) It's the caller's responsibility.
+		(2) Only the caller knows the best way to perform such a feat.
+	There's no guarantee two distinct rows might not get the same offset, so the only
+	reliable check-value is the row number. BISON uses the column number instead,
+	but it also enforces distinct offsets for all rows.
+	"""
+	offset, size = first_fit_decreasing(exceptions, allow_negative=True)
+	check = [-1]*size
+	value = [0]*size
+	for row_id, row_dict in enumerate(exceptions):
+		row_offset = offset[row_id]
+		for column, entry in row_dict.items():
+			index = row_offset + column
+			assert check[index] == -1
+			check[index] = row_id
+			value[index] = entry
+	return {'offset': offset, 'check': check, 'value':value}
+
+###
+###  Major supporting themes: this is where the alleged "deep magic" lives.
+###
+
+def compress_dfa_delta_function(matrix:list) -> dict:
 	"""
 	Per issue #8: A possibly-novel approach to condensing large scanner tables.
 	The method is described at https://github.com/kjosib/booze-tools/issues/8
@@ -43,101 +177,34 @@ def compress_dfa_matrix(*, initial:dict, matrix:list, final:dict) -> dict:
 		residue.append([look_up.get(v) for v in row])
 		exceptions.append({c:x for c,x in enumerate(row) if x not in common})
 	# Compacting the residue plane (a partially-null boolean matrix) by means of equivalence classification:
-	column_class, minimal_columns = find_row_equivalence(zip(*residue), None)
-	residue = list(zip(*minimal_columns))
-	for i, row in enumerate(residue): # Taking care to keep rows sparse:
-		if sum(map(bool, row)) * 2 > len(row):
-			zeros[i], ones[i], residue[i] = ones[i], zeros[i], tuple(None if x is None else int(not x) for x in row)
+	column_class, residue = find_column_equivalence(residue, None)
+	for i in heavy_rows(residue):
+		zeros[i], ones[i] = ones[i], zeros[i]
+		residue[i] = lighten(residue[i])
 	row_class, residue = find_row_equivalence(residue, None)
 	# Converting the final residue to a set-as-displacement-table representation:
-	indices = [[i for i,x in enumerate(row) if x] for row in residue]
-	offset, size = first_fit_decreasing(indices, allow_negative=True)
-	check = [-1]*size
-	for i, (row, base) in enumerate(zip(indices, offset)):
-		for col in row:
-			check[base+col] = i
+	offset, check = encode_boolean_field_as_offset_check(residue)
 	# Tie a pretty bow around it:
-	delta = {
+	return {
 		'exceptions': encode_displacement_function(exceptions),
-		'background': {
+		'bg': {
 			# For most states, the most common entry is the error transition. Thus: exceptions to that rule:
 			'zero': list(zip(*[(index, value) for index, value in enumerate(zeros) if value != -1])),
 			'one': ones,
-			'row_class': row_class, 'column_class': column_class,
+			'row_class': row_class, 'col_class': column_class,
 			'offset': offset, 'check': check,
 		},
 	}
+
+def compress_scanner(*, initial:dict, matrix:list, final:dict) -> dict:
 	height, width = len(matrix), len(matrix[0])
+	delta = compress_dfa_delta_function(matrix)
 	metric = measure_approximate_cost(delta)
-	print("DFA matrix was %d rows, %d cols = %d cells; compact form is about %d cells (%0.2f%%)"%(
+	if VERBOSE: print("DFA matrix was %d rows, %d cols = %d cells; compact form is about %d cells (%0.2f%%)"%(
 		height, width, height * width, metric, (100.0*metric/(height*width))
 	))
 	return {'delta': delta, 'initial': initial, 'final': list(final.keys()), 'rule': list(final.values()),}
 
-def measure_approximate_cost(structure):
-	""" Various bits estimate the size of the structures they return. This makes that consistent. """
-	if isinstance(structure, (list, tuple)): return 1 + sum(map(measure_approximate_cost, structure))
-	elif isinstance(structure, dict): return len(structure) + sum(map(measure_approximate_cost, structure.values()))
-	elif isinstance(structure, int) or structure is None: return 1
-	else: assert False, type(structure)
-
-def first_fit_decreasing(indices: list, *, allow_negative:bool):
-	"""
-	(Aho and Ulman [2]) and (Ziegler [7]) advocate this scheme...
-	This function finds a set of displacements such that no two index[i][j]+displacement[i] are the same.
-	The entry in list `indices` may be a list of columns within a given row which are considered to
-	have legitimate data in some particular sparse matrix. But note that in Python, dictionaries iterate
-	as their keys, so there's a shortcut...
-
-	The result can be used to populate a displacement table, which is a simple kind of perfect-hash.
-		* A hash is "perfect" when it requires at most one probe to determine
-		the presence or absence of a key; in general these must be precomputed.
-		* A hash is "minimal" when (in the lingo of hashing) all the buckets are used.
-		* The holy grail of "minimal perfect hashing" generally requires a more complex function.
-	
-	Rows are considered in decreasing order of density, so if Zipf's law applies then the result
-	will tend to be packed fairly densely: i.e. minimal or nearly so. This tends to be true of the
-	way this particular function gets used in this library.
-	
-	The size of the necessary resulting vector is also returned: it would have to be determined again
-	in every reasonable case otherwise.
-	"""
-	used = set()
-	def first_fit(row: list) -> int:
-		if not row: return 0
-		offset = 0 - min(row) if allow_negative else 0
-		while any(r + offset in used for r in row): offset += 1
-		used.update(r + offset for r in row)
-		return offset
-	
-	displacements = [0] * len(indices)
-	for i in foundation.grade(list(map(len, indices)), descending=True):
-		displacements[i] = first_fit(indices[i])
-	size = max(used)+1 if used else 0
-	print("First-Fit Decreasing placed %d entries among %d positions."%(len(used), size))
-	return displacements, size
-
-def encode_displacement_function(exceptions:list):
-	"""
-	This builds a typical textbook-style displacement table: a simple kind of perfect-hash,
-	which may not be minimal but it's usually decently close.
-	:param exceptions: a list of dictionaries; keys are column numbers, values are destination state numbers.
-	No attempt is made to coalesce similar or identical rows here:
-		(1) It's the caller's responsibility.
-		(2) Only the caller knows the best way to perform such a feat.
-	"""
-	offset, size = first_fit_decreasing(exceptions, allow_negative=True)
-	check = [-1]*size
-	value = [0]*size
-	for row_id, row_dict in enumerate(exceptions):
-		row_offset = offset[row_id]
-		for column, entry in row_dict.items():
-			index = row_offset + column
-			assert check[index] == -1
-			check[index] = row_id
-			value[index] = entry
-	return {'offset': offset, 'check': check, 'value':value}
-
 def decompose_by_default_reduction(matrix:list, essential_errors:set, recovering_states):
 	"""
 	The first phase of compacting a parser's "ACTION" table.
@@ -165,70 +232,82 @@ def decompose_by_default_reduction(matrix:list, essential_errors:set, recovering
 	] for q, (row, default) in enumerate(zip(matrix, reduce))]
 	return reduce, residue
 
-def decompose_by_edit_distance(matrix):
+def decompose_by_edit_distance(action:List[List[int]], focus:List[Set[int]]):
 	"""
-	Note that you CAN get a value of `None` in an 'edit' resulting from this function.
-	Since null means "use the default reduction" in this context, that's considered OK.
-	In a bare-metal implementation, we might decide actually to establish a "rule" for
-	the error response activity: it would therefore get its own number (perhaps -1)
-	and the associated code would be responsible for any error processing. That would
-	take away some of the special-casing associated with parse errors: you just need
-	to make sure the correct code reference is installed in the rule response table.
-	
 	This is one area where performance did become a real problem with a simple approach.
 	Therefore, I've thought about things and come up with something I can be reasonably
 	proud of. All hail the well-considered loop invariant!
+	
+	The basic idea is:
+		1. The outer loop works subject rows in order of _IN_creasing "population".
+		2. The inner loop compares against (already-solved) rows in _DE_creasing population.
+		3. (Critical) An arithmetic test shows when the inner loop becomes hopeless.
+	
+	There are a thousand subtle ways to get this wrong. I've tried most of them.
+	The trick to getting it right is stating the correct post-condition in terms
+	that are easy to verify. Here's the latest try:
+	
+	The edit-chain from R covers at least the columns in `focus[R]` and matches `action[R]`.
+	A useful intermediate datum is which cells are covered by some fallback chain.
 	"""
-	height, width = len(matrix), len(matrix[0])
-	population = [width - row.count(None) for row in matrix]
-	schedule = foundation.grade(population)
-	fallback = [-1] * height
-	# Invariant:
-	# 1. fallback array gets the right information as we go along.
-	# 2. schedule[:bound] mentions no two identical rows, and otherwise mentions rows
-	#    in order of increasing population.
-	# Postcondition: fallback array has the best information, and bound==len(schedule)
-	# This calls for some custom-crafted loop control:
-	bound = 0
-	while bound < len(schedule):
-		q = schedule[bound]
-		metric = population[q] # i.e. the cost of writing out the whole row.
-		x = bound    # We'll iterate DOWNWARD through the schedule, because that's the direction
-		while x > 0: # from which a completely equivalent row is soonest to be found.
-			x -= 1
-			j = schedule[x]
-			if metric + population[j] < population[q]: # then no sparser row can possibly improve on what we have.
-				bound += 1
-				break
-			edit_distance = sum( a != b for a, b in zip(matrix[q], matrix[j]))
-			if edit_distance < metric: metric, fallback[q] = edit_distance, j
-			if edit_distance == 0:  # This row is identical to a previous one.
-				schedule.pop(bound) # Remove this duplicate from consideration henceforth,
-				break               # and look no further for improvement on it.
-		else: bound += 1
-	# Now that we have our fallback vector computed, it's straightforward to work out the edits:
-	def edits(row, basis) -> dict:
-		if basis < 0: return {c:v for c,v in enumerate(row) if v is not None}
-		else: return {c:v for c,(x,v) in enumerate(zip(matrix[basis], row)) if x!=v}
-	return fallback, [edits(row, basis) for row, basis in zip(matrix, fallback)]
+	
+	def find_edits(subject, basis, must_cover, basis_covers):
+		# If the basis covers it WRONGLY, it must be an edit.
+		# If it must be covered and the basis does not, it's an edit.
+		# But if the basis covers it correctly, then it's not an edit.
+		return [
+			c for c in must_cover|basis_covers
+			if not (c in basis_covers and basis[c] == subject[c])
+		]
+	
+	def find_best_fallback():
+		subject, must_cover = action[current], focus[current]
+		fb, best = -1, {c:subject[c] for c in must_cover}
+		for size, candidate in reversed(solved):
+			if len(best) + size < len(must_cover): break
+			golf = find_edits(subject, action[candidate], must_cover, focus[candidate])
+			if len(golf) < len(best): fb,best = candidate,golf
+		must_cover.update(focus[fb])
+		return fb, {c:subject[c] for c in best}
+	
+	assert len(action) == len(focus)
+	fallback = [-1] * len(action)
+	edits = [None] * len(action)
+	solved = []
+	for current in foundation.grade(list(map(len, focus))):
+		fallback[current], edits[current] = find_best_fallback()
+		if edits[current]: bisect.insort_right(solved, (len(focus[current]), current))
+	return fallback, edits
 
-def compress_action_table(matrix:list, essential_errors:set, recovering_states) -> dict:
+def compress_action_table(action:List[List[int]], nonassoc_errors:set) -> dict:
 	"""
 	Produce a compact representation of the "ACTION" table for a typical shift-reduce parser.
-	:param matrix: list-of-lists of parse actions: positive numbers are shifts; negative are reductions, zero is error.
-	:param essential_errors: set of pairs of (state_id, terminal_id) which MUST NOT go to a default reduction.
+	:param action: matrix of parse actions: positive numbers are shifts; negative are reductions, zero is error.
+	:param nonassoc_errors:
+	    set of pairs of (state_id, terminal_id) which must error on account of
+	    non-associativity declarations. They are important for two reasons:
+	    1. Mentioned states must not become "interactive".
+	    2. These error-cells are likely to be isolated, and so good candidates
+	       for the reduction-plane to treat as "don't-care": they'll be picked
+	       up in the shift-plane anyway.
 	:return: a compact structure.
 	"""
-	height, width = len(matrix), len(matrix[0]) # Stats for comparison to the compression method
-	reduce, residue = decompose_by_default_reduction(matrix, essential_errors, recovering_states)
-	fallback, edits = decompose_by_edit_distance(residue)
+	height, width = len(action), len(action[0]) # Stats for comparison to the compression method
+	
+	reduce = compose_reduce_plane(action, nonassoc_errors)
+	model = expansion.parser_reduce_function(**reduce)
+	focus = [ # A boolean map of where the model gets it wrong.
+		set(c for c,i in enumerate(row) if model(r,c) != i)
+		for r,row in enumerate(action)
+	]
+	fallback, edits = decompose_by_edit_distance(action, focus)
+	for q,t in nonassoc_errors:
+		if fallback[q] == -1 and not edits[q]:
+			fallback[q] = -2 # By the way, is this even possible?
 	edit_table = encode_displacement_function(edits)
-	for q, (f, e) in enumerate(zip(fallback, edits)):
-		if f == -1 and not e: # Think about it. This EXACTLY selects default-reduce-only states.
-			edit_table['offset'][q] = len(edit_table['check']) # This marks it as an interactive state.
 	result = {'reduce': reduce, 'fallback': fallback, 'edits': edit_table,}
 	metric = measure_approximate_cost(result)
-	print("Action matrix was %d * %d = %d; compressed to %d (%0.2f%%)"%(
+	if VERBOSE: print("Action matrix was %d * %d = %d; compressed to %d (%0.2f%%)"%(
 		height, width, height*width, metric, (100.0*metric)/(height*width)
 	))
 	return result
@@ -276,7 +355,7 @@ def compress_goto_table(goto_table:list) -> dict:
 	col_class, minimal_columns = find_row_equivalence(zip(*minimal_rows), 0)
 	
 	# Try to figure the ideal column-class ordering to make it fit better:
-	col_class_offset = foundation.everted(foundation.grade([vector.count(0) for vector in minimal_columns]))
+	col_class_offset = foundation.collation([vector.count(0) for vector in minimal_columns])
 	minimal_rows = [{col_class_offset[c]:x for c, x in enumerate(row) if x} for row in zip(*minimal_columns)]
 	
 	# Build a single "residue vector" and row-offsets for efficient packing of the residue.
@@ -293,27 +372,47 @@ def compress_goto_table(goto_table:list) -> dict:
 		col_index[nonterm_id] = col_class_offset[cls_id] + hi_water_mark
 	
 	# Wrap up and return.
-	print("GOTO table original size: %d rows, %d columns -> %d cells"%(height, width, height * width))
+	if VERBOSE: print("GOTO table original size: %d rows, %d columns -> %d cells"%(height, width, height * width))
 	result = {'row_index': row_index, 'col_index': col_index, 'quotient': quotient+residue, 'mark': hi_water_mark}
 	metric = measure_approximate_cost(result)
-	print("GOTO compact size: %d (%.2f%%)"%(metric, 100.0*metric/(height * width)))
+	if VERBOSE: print("GOTO compact size: %d (%.2f%%)"%(metric, 100.0*metric/(height * width)))
 	return result
 
-def find_row_equivalence(matrix, do_not_care):
+def compose_reduce_plane(action, nonassoc_errors):
 	"""
-	This typically gets called twice per matrix with a transposition for the columns.
-	:param matrix: A rank-two tensor. Iterable of rows.
-	:param do_not_care: Cells with this value may be assigned any convenient value for the purpose.
+	What's going on here?
+	I want to try to retain the error entries, even in states with a default reduction.
+	The glimmer of hope centers on a semantic partitioning of the ACTION table.
 	"""
-	index, classes = [], []
-	for row in matrix:
-		for class_id, candidate_class in enumerate(classes):
-			if all(a == b or do_not_care in (a, b) for a, b in zip(row, candidate_class)):
-				index.append(class_id)
-				for c, value in enumerate(row):
-					if value != do_not_care: candidate_class[c] = value
-				break
-		else:
-			index.append(foundation.allocate(classes, list(row)))
-	return index, classes
-
+	
+	def decide(i, dr):
+		if i == 0: return False
+		if i == dr: return True
+		return None
+	
+	def best_d_reduce(row):
+		x = [a for a in row if a < 0]
+		if x:
+			i, count = collections.Counter(x).most_common(1)[0]
+			if count > 1: return i
+		return 0
+	
+	d_reduce = list(map(best_d_reduce, action))  # Default Reduction Table
+	r_plane = [
+		[decide(i, dr) for i in row]
+		for row, dr in zip(action, d_reduce)
+	]
+	for q,t in nonassoc_errors: r_plane[q][t] = None # These are likely to be one-offs.
+	col_class, minimal_cols = find_row_equivalence(zip(*r_plane), None)
+	midpoint = list(zip(*minimal_cols))
+	# hot = compaction.heavy_rows(midpoint)
+	# for r in hot: midpoint[r] = compaction.lighten(midpoint[r])
+	row_class, minimal_rows = find_row_equivalence(midpoint, None)
+	offset, check = encode_boolean_field_as_offset_check(minimal_rows)
+	return {
+		'd_reduce': d_reduce,
+		'row_class': row_class,
+		'col_class': col_class,
+		'offset': offset,
+		'check': check,
+	}
